@@ -27,17 +27,64 @@ configuration file can be triggered with a `kill -10 <dispatcher pid>`.
 """
 
 import argparse
+import logging
+import logging.config
+import logging.handlers
+import os
 import sys
 
+import yaml
+
 from trollmoves.dispatcher import Dispatcher
-from trollmoves.logging import setup_logging, add_logging_options_to_parser
+
+LOG_FORMAT = "[%(asctime)s %(levelname)-8s] %(message)s"
+logger = logging.getLogger('dispatcher')
+
+log_levels = {
+    0: logging.WARN,
+    1: logging.INFO,
+    2: logging.DEBUG,
+}
 
 
-def parse_args():
-    """Parse commandline arguments."""
+def setup_logging(cmd_args):
+    """Set up logging."""
+    if cmd_args.log_config is not None:
+        with open(cmd_args.log_config) as fd:
+            log_dict = yaml.safe_load(fd.read())
+            logging.config.dictConfig(log_dict)
+            return
+
+    root = logging.getLogger('')
+    root.setLevel(log_levels[cmd_args.verbosity])
+
+    if cmd_args.log:
+        fh_ = logging.handlers.TimedRotatingFileHandler(
+            os.path.join(cmd_args.log),
+            "midnight",
+            backupCount=7)
+    else:
+        fh_ = logging.StreamHandler()
+
+    formatter = logging.Formatter(LOG_FORMAT)
+    fh_.setFormatter(formatter)
+
+    root.addHandler(fh_)
+
+
+def main():
+    """Start and run the dispatcher."""
     parser = argparse.ArgumentParser()
     parser.add_argument("config_file",
                         help="The configuration file to run on.")
+    parser.add_argument("-l", "--log",
+                        help="The file to log to. stdout otherwise.")
+    parser.add_argument("-c", "--log-config",
+                        help="Log config file to use instead of the standard logging.")
+    parser.add_argument("-v", "--verbose", dest="verbosity", action="count", default=0,
+                        help="Verbosity (between 1 and 2 occurrences with more leading to more "
+                        "verbose logging). WARN=0, INFO=1, "
+                        "DEBUG=2. This is overridden by the log config file if specified.")
     parser.add_argument(
         "-p", "--publish-port", type=int, dest="pub_port", nargs='?',
         const=0, default=None,
@@ -46,14 +93,8 @@ def parse_args():
     parser.add_argument("-n", "--publish-nameserver", nargs='*',
                         dest="pub_nameservers",
                         help="Nameserver for publisher to connect to")
-    add_logging_options_to_parser(parser, legacy=True)
-    return parser.parse_args()
-
-
-def main():
-    """Start and run the dispatcher."""
-    cmd_args = parse_args()
-    logger = setup_logging("dispatcher", cmd_args)
+    cmd_args = parser.parse_args()
+    setup_logging(cmd_args)
     logger.info("Starting up.")
 
     try:
@@ -64,7 +105,8 @@ def main():
         logger.error('Dispatcher crashed: %s', str(err))
         sys.exit(1)
     try:
-        dispatcher.run()
+        dispatcher.start()
+        dispatcher.join()
     except KeyboardInterrupt:
         logger.debug("Interrupting")
     finally:
