@@ -22,31 +22,32 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 """Trollmoves client."""
-import argparse
+
 import logging
 import os
 import socket
+import sys
 import time
+import tarfile
 from collections import deque
-from configparser import ConfigParser
+from six.moves.configparser import RawConfigParser
 from threading import Lock, Thread, Event
 import hashlib
-from urllib.parse import urlparse, urlunparse
+import six
+from six.moves.urllib.parse import urlparse, urlunparse
 import subprocess
 from contextlib import suppress
-
-import tarfile
+import pyinotify
 from zmq import LINGER, POLLIN, REQ, Poller
 import bz2
+
 from posttroll import get_context
 from posttroll.message import Message, MessageError
-from posttroll.publisher import create_publisher_from_dict_config
+from posttroll.publisher import NoisyPublisher
 from posttroll.subscriber import Subscriber
 from trollsift.parser import compose
 
 from trollmoves import heartbeat_monitor
-from trollmoves.logging import add_logging_options_to_parser
-from trollmoves.move_it_base import MoveItBase
 from trollmoves.utils import get_local_ips
 from trollmoves.utils import gen_dict_extract, translate_dict
 from trollmoves.movers import CTimer
@@ -62,7 +63,6 @@ ongoing_hot_spare_timers = dict()
 
 DEFAULT_REQ_TIMEOUT = 1
 SERVER_HEARTBEAT_TOPIC = "/heartbeat/move_it_server"
-CLIENT_HEARTBEAT_TOPIC_BASE = "/heartbeat/move_it"
 
 COMPRESSED_ENDINGS = {'xrit': ['C_'],
                       'tar': ['.tar', '.tar.gz', '.tgz', '.tar.bz2'],
@@ -72,117 +72,77 @@ BUNZIP_BLOCK_SIZE = 1024
 LISTENER_CHECK_INTERVAL = 1
 
 
-def is_localhost(host):
-    """Check if host is localhost."""
-    return socket.gethostbyname(host) in get_local_ips()
-
-
+# Config management
 def read_config(filename):
     """Read the config file called *filename*."""
-    cp_ = ConfigParser(interpolation=None)
-    with open(filename) as config_file:
-        cp_.read_file(config_file)
+    cp_ = RawConfigParser()
+    cp_.read(filename)
 
     res = {}
 
     for section in cp_.sections():
         res[section] = dict(cp_.items(section))
-        _set_config_defaults(res[section])
-        _parse_boolean_config_items(res[section], cp_[section])
-        _parse_nameservers(res[section], cp_[section])
-        if not _check_provider_config(res, section):
+        res[section].setdefault("delete", False)
+        if res[section]["delete"] in ["", "False", "false", "0", "off"]:
+            res[section]["delete"] = False
+        if res[section]["delete"] in ["True", "true", "on", "1"]:
+            res[section]["delete"] = True
+        res[section].setdefault("working_directory", None)
+        res[section].setdefault("compression", False)
+        res[section].setdefault("xritdecompressor", None)
+        res[section].setdefault("heartbeat", True)
+        res[section].setdefault("req_timeout", DEFAULT_REQ_TIMEOUT)
+        res[section].setdefault("transfer_req_timeout", 10 * DEFAULT_REQ_TIMEOUT)
+        res[section].setdefault("nameservers", None)
+        if res[section]["heartbeat"] in ["", "False", "false", "0", "off"]:
+            res[section]["heartbeat"] = False
+
+        if "providers" not in res[section]:
+            LOGGER.warning("Incomplete section %s: add an 'providers' item.",
+                           section)
+            LOGGER.info("Ignoring section %s: incomplete.",
+                        section)
+            del res[section]
             continue
-        if not _check_destination(res, section):
+        else:
+            res[section]["providers"] = [
+                "tcp://" + item.split('/', 1)[0] for item in res[section]["providers"].split()
+            ]
+
+        if "destination" not in res[section]:
+            LOGGER.warning("Incomplete section %s: add an 'destination' item.",
+                           section)
+            LOGGER.info("Ignoring section %s: incomplete.", section)
+            del res[section]
             continue
-        if not _check_subscribing(res, section):
+
+        if "topic" in res[section]:
+            try:
+                res[section]["publish_port"] = int(res[section][
+                    "publish_port"])
+            except (KeyError, ValueError):
+                res[section]["publish_port"] = 0
+        elif not res[section]["heartbeat"]:
+            # We have no topics and therefor no subscriber (if you want to
+            # subscribe everything, then explicit specify an empty topic).
+            LOGGER.warning("Incomplete section %s: add an 'topic' "
+                           "item or enable heartbeat.", section)
+            LOGGER.info("Ignoring section %s: incomplete.", section)
+            del res[section]
             continue
 
     return res
 
 
-def _set_config_defaults(conf):
-    conf.setdefault("delete", False)
-    conf.setdefault("working_directory", None)
-    conf.setdefault("compression", False)
-    conf.setdefault("xritdecompressor", None)
-    conf.setdefault("heartbeat", True)
-    conf.setdefault("req_timeout", DEFAULT_REQ_TIMEOUT)
-    conf.setdefault("transfer_req_timeout", 10 * DEFAULT_REQ_TIMEOUT)
-    conf.setdefault("nameservers", None)
-    conf.setdefault("create_target_directory", True)
-
-
-def _parse_boolean_config_items(conf, raw_conf):
-    for key in ["delete", "heartbeat", "create_target_directory"]:
-        try:
-            val = raw_conf.getboolean(key)
-        except ValueError:
-            continue
-        if val is not None:
-            conf[key] = val
-
-
-def _parse_nameservers(conf, raw_conf):
-    try:
-        val = raw_conf.getboolean("nameservers")
-    except ValueError:
-        val = conf["nameservers"]
-    if isinstance(val, str):
-        val = val.split()
-    conf["nameservers"] = val
-
-
-def _check_provider_config(conf, section):
-    if "providers" not in conf[section]:
-        LOGGER.warning("Incomplete section %s: add an 'providers' item.",
-                       section)
-        LOGGER.info("Ignoring section %s: incomplete.",
-                    section)
-        del conf[section]
-        return False
-
-    conf[section]["providers"] = [
-        "tcp://" + item.split('/', 1)[0] for item in conf[section]["providers"].split()
-    ]
-    return True
-
-
-def _check_subscribing(res, section):
-    if "topic" in res[section]:
-        try:
-            res[section]["publish_port"] = int(res[section][
-                "publish_port"])
-        except (KeyError, ValueError):
-            res[section]["publish_port"] = 0
-    elif not res[section]["heartbeat"]:
-        # We have no topics and therefor no subscriber (if you want to
-        # subscribe everything, then explicit specify an empty topic).
-        LOGGER.warning("Incomplete section %s: add an 'topic' "
-                       "item or enable heartbeat.", section)
-        LOGGER.info("Ignoring section %s: incomplete.", section)
-        del res[section]
-        return False
-    return True
-
-
-def _check_destination(res, section):
-    if "destination" not in res[section]:
-        LOGGER.warning("Incomplete section %s: add an 'destination' item.",
-                       section)
-        LOGGER.info("Ignoring section %s: incomplete.", section)
-        del res[section]
-        return False
-    return True
-
-
 class Listener(Thread):
     """PyTroll listener class for reading messages for Trollduction."""
 
-    def __init__(self, address, topics, *args, die_event=None, **kwargs):
+    def __init__(self, address, topics, callback, *args, die_event=None, **kwargs):
         """Init Listener object."""
-        super().__init__()
+        super(Listener, self).__init__()
 
         self.topics = topics
+        self.callback = callback
         self.subscriber = None
         self.address = address
         self.running = False
@@ -196,7 +156,7 @@ class Listener(Thread):
     def restart(self):
         """Restart the listener, returns a new running instance."""
         self.stop()
-        new_listener = self.__class__(self.address, self.topics, *self.cargs,
+        new_listener = self.__class__(self.address, self.topics, self.callback, *self.cargs,
                                       die_event=self.die_event, **self.ckwargs)
         new_listener.death_count = self.death_count + 1
         new_listener.start()
@@ -215,67 +175,68 @@ class Listener(Thread):
         """Run listener."""
         try:
             with heartbeat_monitor.Monitor(self.restart_event, **self.ckwargs) as beat_monitor:
+
                 self.running = True
+
                 while self.running:
+                    # Loop for restart.
+
                     LOGGER.debug("Starting listener %s", str(self.address))
                     self.create_subscriber()
-                    self._get_messages(beat_monitor)
+
+                    for msg in self.subscriber(timeout=1):
+                        if not self.running:
+                            break
+
+                        # If the heartbeat didn't arrive, restart the subscriber
+                        if self.restart_event.is_set():
+                            LOGGER.warning("Missing a heartbeat, restarting the subscriber to %s.",
+                                           str(self.subscriber.addresses))
+                            self.restart_event.clear()
+                            self.stop()
+                            self.running = True
+                            break
+
+                        if msg is None:
+                            continue
+
+                        LOGGER.debug("Receiving (SUB) %s", str(msg))
+
+                        beat_monitor(msg)
+                        if msg.type == "beat":
+                            self.death_count = 0
+                            continue
+                        # Handle public "push" messages as a hot spare client
+                        if msg.type == "push":
+                            # TODO: these need to be checked and acted if
+                            # the transfers are not finished on primary
+                            # client and are not cleared
+                            LOGGER.debug("Primary client published 'push'")
+                            add_to_ongoing(msg)
+
+                        # Handle public "ack" messages as a hot spare client
+                        if msg.type == "ack":
+                            LOGGER.debug("Primary client finished transfer")
+                            _ = add_to_file_cache(msg)
+                            _ = clean_ongoing_transfer(get_msg_uid(msg))
+
+                        # If this is a hot spare client, wait for a while
+                        # for a public "push" message which will update
+                        # the ongoing transfers before starting processing here
+                        delay = self.ckwargs.get("processing_delay", False)
+                        if delay:
+                            add_timer(float(delay), self.callback, msg, *self.cargs,
+                                      **self.ckwargs)
+                        else:
+                            self.callback(msg, *self.cargs, **self.ckwargs)
+
+                    LOGGER.debug("Exiting listener %s", str(self.address))
         except Exception as err:
             LOGGER.exception("Listener died.")
             self.cause_of_death = err
             with suppress(AttributeError):
                 self.die_event.set()
 
-    def _get_messages(self, beat_monitor):
-        for msg in self.subscriber(timeout=1):
-            if not self.running:
-                break
-            if not self._check_heartbeat():
-                break
-            if msg is None:
-                continue
-
-            LOGGER.debug("Receiving (SUB) %s", str(msg))
-
-            beat_monitor(msg)
-
-            if self._is_message_already_handled(msg):
-                continue
-
-            self._process_message(msg)
-
-        LOGGER.debug("Exiting listener %s", str(self.address))
-
-    def _check_heartbeat(self):
-        if self.restart_event.is_set():
-            LOGGER.warning("Missing a heartbeat, restarting the subscriber to %s.",
-                           str(self.subscriber.addresses))
-            self.restart_event.clear()
-            self.stop()
-            self.running = True
-            return False
-        return True
-
-    def _is_message_already_handled(self, msg):
-        return (self._handle_beat_message(msg) or _handle_push_message(msg) or
-                _handle_ack_message(msg) or _handle_message_from_another_client(msg))
-
-    def _handle_beat_message(self, msg):
-        if msg.type == "beat":
-            self.death_count = 0
-            return True
-        return False
-
-    def _process_message(self, msg):
-        delay = self.ckwargs.get("processing_delay", False)
-        if delay:
-            # If this is a hot spare client, wait for a while
-            # for a public "push" message which will update
-            # the ongoing transfers before starting processing here
-            add_request_push_timer(float(delay), msg, *self.cargs, **self.ckwargs)
-        else:
-            request_push(msg, *self.cargs, **self.ckwargs)
-
     def stop(self):
         """Stop subscriber and delete the instance."""
         self.running = False
@@ -285,36 +246,6 @@ class Listener(Thread):
             self.subscriber = None
 
 
-def _handle_push_message(msg):
-    if msg.type == "push":
-        # TODO: these need to be checked and acted if
-        # the transfers are not finished on primary
-        # client and are not cleared
-        LOGGER.debug("Primary client published 'push'")
-        add_to_ongoing_transfers(msg)
-        return True
-    return False
-
-
-def _handle_ack_message(msg):
-    if msg.type == "ack":
-        LOGGER.debug("Primary client finished transfer")
-        _ = add_to_file_cache(msg)
-        _ = clean_ongoing_transfer(get_msg_uid(msg))
-        return True
-    return False
-
-
-def _handle_message_from_another_client(msg):
-    if msg.type == "file" and "request_address" not in msg.data:
-        LOGGER.debug("Ignoring 'file' message from primary client.")
-        add_to_ongoing_transfers(msg)
-        _ = add_to_file_cache(msg)
-        _ = clean_ongoing_transfer(get_msg_uid(msg))
-        return True
-    return False
-
-
 def clean_ongoing_transfer(uid):
     """Clear transfer for the given UID from the cache."""
     with ongoing_transfers_lock:
@@ -409,7 +340,7 @@ def resend_if_local(msg, publisher):
     """Resend the message provided all uris point to local files."""
     for uri in gen_dict_extract(msg.data, 'uri'):
         urlobj = urlparse(uri)
-        if not publisher or not is_localhost(urlobj.netloc):
+        if not publisher or not socket.gethostbyname(urlobj.netloc) in get_local_ips():
             return
 
     LOGGER.debug('Sending: %s', str(msg))
@@ -437,8 +368,6 @@ def create_push_req_message(msg, destination, login):
 def create_local_dir(destination, local_root, mode=0o777):
     """Create the local directory if it doesn't exist and return that path."""
     duri = urlparse(destination)
-    if duri.scheme in ('s3'):
-        return None
     local_dir = os.path.join(*([local_root] + duri.path.split(os.path.sep)))
 
     if not os.path.exists(local_dir):
@@ -490,40 +419,32 @@ def unpack_and_create_local_message(msg, local_dir, **kwargs):
 def make_uris(msg, destination, login=None):
     """Create local URIs for the received files."""
     duri = urlparse(destination)
-    scheme = duri.scheme
-    netloc = duri.netloc
-    if scheme != "s3" and empty_or_localhost(duri.hostname):
-        scheme = ""
-        netloc = ""
-    elif netloc:
+    scheme = duri.scheme or 'ssh'
+    dest_hostname = duri.hostname or socket.gethostname()
+    if socket.gethostbyname(dest_hostname) in get_local_ips():
+        scheme_, host_ = "ssh", dest_hostname  # local file
+    else:
+        scheme_, host_ = scheme, dest_hostname  # remote file
         if login:
             # Add (only) user to uri.
-            netloc = login.split(":")[0] + "@" + netloc
+            host_ = login.split(":")[0] + "@" + host_
 
     def uri_callback(var):
         uid = var['uid']
         path = os.path.join(duri.path, uid)
-        var['uri'] = urlunparse((scheme, netloc, path, "", "", ""))
+        var['uri'] = urlunparse((scheme_, host_, path, "", "", ""))
         return var
     msg.data = translate_dict(msg.data, ('uri', 'uid'), uri_callback)
     return msg
 
 
-def empty_or_localhost(hostname):
-    """Check that hostname is either empty or referring to localhost."""
-    return ((not hostname) or (hostname and is_localhost(hostname)))
-
-
 def replace_mda(msg, kwargs):
-    """Replace message metadata with items in kwargs dict."""
+    """Replace messate metadata with itmes in kwargs dict."""
     for key in msg.data:
         if key in kwargs:
-            try:
-                replacement = dict(item.split(':') for item in kwargs[key].split('|'))
-                replacement = replacement[msg.data[key]]
-            except ValueError:
-                replacement = kwargs[key]
-            msg.data[key] = replacement
+            replacement = dict(item.split(':')
+                               for item in kwargs[key].split('|'))
+            msg.data[key] = replacement[msg.data[key]]
     return msg
 
 
@@ -582,34 +503,34 @@ def get_next_msg(uid):
         return ongoing_transfers[uid].pop(0)
 
 
-def add_request_push_timer(timeout, msg, *args, **kwargs):
+def add_timer(timeout, callback, msg, *args, **kwargs):
     """Add a timer for hot spare."""
     huid = get_msg_uid(msg)
     cargs = [msg] + list(args)
     with hot_spare_timer_lock:
-        timer = CTimer(timeout, request_push, args=cargs, kwargs=kwargs)
+        timer = CTimer(timeout, callback, args=cargs, kwargs=kwargs)
         ongoing_hot_spare_timers[huid] = timer
         ongoing_hot_spare_timers[huid].start()
     LOGGER.debug("Added timer for UID %s.", huid)
 
 
-def add_to_ongoing_transfers(msg):
+def add_to_ongoing(msg):
     """Add message to ongoing transfers.
 
-    Return None if similar message was already received, otherwise the hashed uid of the message.
+    Return True if similar message was already received, False otherwise.
     """
-    hashed_uid = get_msg_uid(msg)
+    huid = get_msg_uid(msg)
     with hot_spare_timer_lock:
-        timer = ongoing_hot_spare_timers.pop(hashed_uid, None)
+        timer = ongoing_hot_spare_timers.pop(huid, None)
         if timer is not None:
             timer.cancel()
-            LOGGER.debug("Cleared timer for UID %s.", hashed_uid)
+            LOGGER.debug("Cleared timer for UID %s.", huid)
     with ongoing_transfers_lock:
-        if hashed_uid in ongoing_transfers:
-            ongoing_transfers[hashed_uid].append(msg)
+        if huid in ongoing_transfers:
+            ongoing_transfers[huid].append(msg)
             return None
-        ongoing_transfers[hashed_uid] = [msg]
-        return hashed_uid
+        ongoing_transfers[huid] = [msg]
+        return huid
 
 
 def add_to_file_cache(msg):
@@ -621,51 +542,65 @@ def add_to_file_cache(msg):
                 file_cache.append(uid)
 
 
-def request_push(msg_in, destination, login=None, publisher=None, **kwargs):
+def request_push(msg_in, destination, login=None, sync_publisher=None, publisher=None, **kwargs):
     """Request a push for data."""
-    hashed_uid = add_to_ongoing_transfers(msg_in)
-    if hashed_uid is None:
+    huid = add_to_ongoing(msg_in)
+    if huid is None:
         return
 
     if already_received(msg_in):
         timeout = float(kwargs["req_timeout"])
         send_ack(msg_in, timeout)
-        _ = clean_ongoing_transfer(hashed_uid)
+        _ = clean_ongoing_transfer(huid)
         return
 
-    _request_files(hashed_uid, destination, login, publisher, **kwargs)
+    for msg in iterate_messages(huid):
+        try:
+            _destination = compose(destination, msg.data)
+        except KeyError as ke:
+            LOGGER.error("Format identifier is missing from the msg.data: %s", str(ke))
+            raise
+        except ValueError as ve:
+            LOGGER.error("Type of format identifier doesn't match the type in m msg.data: %s", str(ve))
+            raise
+        except AttributeError as ae:
+            LOGGER.error("msg or msg.data is None: %s", str(ae))
+            raise
+        req, fake_req = create_push_req_message(msg, _destination, login)
+        LOGGER.info("Requesting: %s", str(fake_req))
+        timeout = float(kwargs["transfer_req_timeout"])
+        local_dir = create_local_dir(_destination, kwargs.get('ftp_root', '/'))
 
-
-def _request_files(hashed_uid, destination, login, publisher, **kwargs):
-    for msg in iterate_messages(hashed_uid):
-        _destination = _compose_destination(destination, msg)
-
-        req, no_credentials_req = create_push_req_message(msg, _destination, login)
-        LOGGER.info("Requesting: %s", str(no_credentials_req))
-        if kwargs.get('create_target_directory', True):
-            local_dir = create_local_dir(_destination, kwargs.get('ftp_root', '/'))
-        else:
-            local_dir = None
-
-        publisher.send(str(no_credentials_req))
-
-        response, hostname = send_request(msg, req, float(kwargs["transfer_req_timeout"]))
+        if sync_publisher:
+            sync_publisher.send(str(fake_req))
+        response, hostname = send_request(msg, req, timeout)
 
         if response and response.type in ['file', 'collection', 'dataset']:
             LOGGER.debug("Server done sending file")
             add_to_file_cache(msg)
-            _send_ack_message(msg, publisher)
-
+            if sync_publisher:
+                # Send an 'ack' message so that possible hot spares know
+                # the primary has completed the request
+                msg = Message(msg.subject, 'ack', msg.data)
+                LOGGER.debug("Sending a public 'ack' of completed transfer: %s",
+                             str(msg))
+                sync_publisher.send(str(msg))
             try:
-                lmsg = unpack_and_create_local_message(response, local_dir, **kwargs)
-                lmsg = _update_local_message(lmsg, _destination, login, response, **kwargs)
+                lmsg = unpack_and_create_local_message(response, local_dir,
+                                                       **kwargs)
             except IOError:
                 LOGGER.exception("Couldn't unpack %s", str(response))
                 continue
+            if publisher:
+                lmsg = make_uris(lmsg, _destination, login)
+                lmsg.data['origin'] = response.data['request_address']
+                lmsg.data.pop('request_address', None)
+                lmsg = replace_mda(lmsg, kwargs)
+                lmsg.data.pop('destination', None)
 
-            LOGGER.debug("publishing %s", str(lmsg))
-            publisher.send(str(lmsg))
-            terminate_transfers(hashed_uid, float(kwargs["req_timeout"]))
+                LOGGER.debug("publishing %s", str(lmsg))
+                publisher.send(str(lmsg))
+            terminate_transfers(huid, float(kwargs["req_timeout"]))
             break
         else:
             LOGGER.error("Failed to get valid response from server %s: %s",
@@ -673,42 +608,7 @@ def _request_files(hashed_uid, destination, login, publisher, **kwargs):
     else:
         LOGGER.warning('Could not get a working source for requesting %s',
                        str(msg))
-        terminate_transfers(hashed_uid, float(kwargs["req_timeout"]))
-
-
-def _compose_destination(destination, msg):
-    try:
-        _destination = compose(destination, msg.data)
-    except KeyError as ke:
-        LOGGER.error("Format identifier is missing from the msg.data: %s", str(ke))
-        raise
-    except ValueError as ve:
-        LOGGER.error("Type of format identifier doesn't match the type in m msg.data: %s", str(ve))
-        raise
-    except AttributeError as ae:
-        LOGGER.error("msg or msg.data is None: %s", str(ae))
-        raise
-    return _destination
-
-
-def _send_ack_message(msg, publisher):
-    """Send an 'ack' message.
-
-    This is for the possible hot spare clients so they know the primary has completed the request.
-    """
-    msg = Message(msg.subject, 'ack', msg.data)
-    LOGGER.debug("Sending a public 'ack' of completed transfer: %s", str(msg))
-    publisher.send(str(msg))
-
-
-def _update_local_message(lmsg, _destination, login, response, **kwargs):
-    lmsg = make_uris(lmsg, _destination, login)
-    lmsg.data['origin'] = response.data['request_address']
-    lmsg.data.pop('request_address', None)
-    lmsg = replace_mda(lmsg, kwargs)
-    lmsg.data.pop('destination', None)
-
-    return lmsg
+        terminate_transfers(huid, float(kwargs["req_timeout"]))
 
 
 class Chain(Thread):
@@ -716,44 +616,39 @@ class Chain(Thread):
 
     def __init__(self, name, config):
         """Init a chain object."""
-        super().__init__()
+        super(Chain, self).__init__()
         self._config = config
         self._name = name
         self.publisher = None
-        self._pub_starter = None
         self.listeners = {}
         self.listener_died_event = Event()
         self.running = True
-        self.setup_publisher()
+        self.sync_publisher = None
 
-    def setup_publisher(self):
-        """Initialize publisher."""
-        if self.publisher is None:
-            with suppress(KeyError, NameError):
-                nameservers = self._config["nameservers"]
-                pub_settings = {
-                    "name": "move_it_" + self._name,
-                    "port": self._config["publish_port"],
-                    "nameservers": nameservers,
-                }
-                self._pub_starter = create_publisher_from_dict_config(pub_settings)
-                self.publisher = self._pub_starter.start()
+        # Setup publisher
+        try:
+            nameservers = self._config["nameservers"]
+            if nameservers:
+                nameservers = nameservers.split()
+            self.publisher = NoisyPublisher(
+                "move_it_" + self._name,
+                port=self._config["publish_port"],
+                nameservers=nameservers)
+            self.publisher.start()
+        except (KeyError, NameError):
+            pass
 
-    def setup_listeners(self, keep_providers=None):
+    def setup_listeners(self, callback, sync_publisher):
         """Set up the listeners."""
-        keep_providers = keep_providers or []
+        self.callback = callback
+        self.sync_publisher = sync_publisher
         try:
             topics = []
             if "topic" in self._config:
                 topics.append(self._config["topic"])
             if self._config.get("heartbeat", False):
                 topics.append(SERVER_HEARTBEAT_TOPIC)
-                # Subscribe also to heartbeat messages of other clients
-                topics.append(CLIENT_HEARTBEAT_TOPIC_BASE + '_' + self._name)
             for provider in self._config["providers"]:
-                if provider in keep_providers and provider in self.listeners:
-                    LOGGER.debug("Not restarting Listener to %s, config not changed.", provider)
-                    continue
                 if '/' in provider.split(':')[-1]:
                     parts = urlparse(provider)
                     if parts.scheme != '':
@@ -770,6 +665,8 @@ class Chain(Thread):
                 listener = Listener(
                     provider,
                     topics,
+                    callback,
+                    sync_publisher=sync_publisher,
                     publisher=self.publisher,
                     die_event=self.listener_died_event,
                     **self._config)
@@ -821,90 +718,50 @@ class Chain(Thread):
                 return False
         return True
 
-    def get_unchanged_providers(self, other_config):
-        """Get a list of providers that have not changed between this and other config."""
-        if self._config["topic"] != other_config["topic"]:
-            return []
-        return list(set(self._config["providers"]).intersection(set(other_config["providers"])))
-
-    def publisher_needs_restarting(self, other_config):
-        """Check that current config is the same as `other_config`."""
-        for key in ["nameservers", "publish_port"]:
-            if self._config[key] != other_config[key]:
-                return True
-        return False
-
-    def refresh(self, new_config):
-        """Refresh the chain with new config."""
-        publisher_needs_restarting = self.publisher_needs_restarting(new_config)
-        unchanged_providers = self.get_unchanged_providers(new_config)
-        self._config = new_config
-        if publisher_needs_restarting:
-            self._refresh_publisher()
-        self._refresh_listeners(unchanged_providers)
-        if not self.running:
-            self.start()
-
-    def _refresh_publisher(self):
-        self._stop_publisher()
-        self.setup_publisher()
-
-    def _refresh_listeners(self, unchanged_providers):
-        self.reset_listeners(keep_providers=unchanged_providers)
-        self.setup_listeners(keep_providers=unchanged_providers)
-
-    def reset_listeners(self, keep_providers=None):
+    def reset_listeners(self):
         """Reset the listeners."""
-        keep_providers = keep_providers or []
-        kept_listeners = {}
-        for key, listener in self.listeners.items():
-            if key in keep_providers:
-                kept_listeners[key] = listener
-                continue
+        for listener in self.listeners.values():
             listener.stop()
-        self.listeners = kept_listeners
+        self.listeners = {}
 
     def stop(self):
         """Stop the chain."""
-        self._stop_publisher()
         self.running = False
-        self.reset_listeners()
-
-    def _stop_publisher(self):
         if self.publisher:
-            self._pub_starter.stop()
-            self._pub_starter = None
-            self.publisher = None
+            self.publisher.stop()
+        self.reset_listeners()
 
     def restart(self):
         """Restart the chain, return a new running instance."""
         self.stop()
         new_chain = self.__class__(self._name, self._config)
-        new_chain.setup_listeners()
+        new_chain.setup_listeners(self.callback, self.sync_publisher)
         new_chain.start()
         return new_chain
 
 
-def reload_config(filename, chains):
+def reload_config(filename, chains, callback=request_push, sync_publisher=None):
     """Rebuild chains if needed (if the configuration changed) from *filename*."""
     LOGGER.debug("New config file detected: %s", filename)
 
     new_configs = read_config(filename)
 
     # setup new chains
+
     for key, new_config in new_configs.items():
         if key in chains:
             if chains[key].config_equals(new_config):
                 continue
+
+            chains[key].stop()
             verb = "Updated"
             LOGGER.debug("Updating %s", key)
         else:
             verb = "Added"
             LOGGER.debug("Adding %s", key)
-            chains[key] = Chain(key, new_config)
-            chains[key].start()
-
-        chains[key].refresh(new_config)
+        chains[key] = Chain(key, new_config)
+        chains[key].setup_listeners(callback, sync_publisher)
+        chains[key].start()
 
         LOGGER.debug("%s %s", verb, key)
 
@@ -919,7 +776,7 @@ def reload_config(filename, chains):
     LOGGER.debug("Reloaded config from %s", filename)
 
 
-class PushRequester:
+class PushRequester(object):
     """Base requester class."""
 
     request_retries = 3
@@ -939,6 +796,7 @@ class PushRequester:
     def connect(self):
         """Connect to the server."""
         self._socket = get_context().socket(REQ)
+        # self._socket.setsockopt(zmq.RCVTIMEO, 500)  # milliseconds
         self._socket.connect(self._reqaddress)
         self._poller.register(self._socket, POLLIN)
 
@@ -1010,43 +868,53 @@ class PushRequester:
 
         return rep
 
-
-def parse_args(args=None):
-    """Parse commandline arguments."""
-    parser = argparse.ArgumentParser()
-    parser.add_argument("config_file",
-                        help="The configuration file to run on.")
-    add_logging_options_to_parser(parser, legacy=True)
-    return parser.parse_args(args)
+# Generic event handler
 
 
-class MoveItClient(MoveItBase):
-    """Trollmoves client class."""
+class EventHandler(pyinotify.ProcessEvent):
+    """Handle events with a generic *fun* function."""
 
-    def __init__(self, cmd_args):
-        """Initialize client."""
-        self.name = "move_it_client"
-        super().__init__(cmd_args)
+    def __init__(self, fun, *args, **kwargs):
+        """Initialize handler."""
+        pyinotify.ProcessEvent.__init__(self, *args, **kwargs)
+        self._fun = fun
 
-    def reload_cfg_file(self, filename, *args, **kwargs):
-        """Reload configuration file."""
-        reload_config(filename, self.chains, *args, **kwargs)
+    def process_IN_CLOSE_WRITE(self, event):
+        """Process on closing after writing."""
+        self._fun(event.pathname)
 
-    def signal_reload_cfg_file(self, *args):
-        """Handle reload signal."""
-        reload_config(self.cmd_args.config_file, self.chains,
-                      publisher=self.publisher)
+    def process_IN_CREATE(self, event):
+        """Process on closing after linking."""
+        try:
+            if os.stat(event.pathname).st_nlink > 1:
+                self._fun(event.pathname)
+        except OSError:
+            return
 
-    def _run(self):
-        for chain_name in self.chains:
-            if not self.chains[chain_name].is_alive():
-                self.chains[chain_name] = self.chains[chain_name].restart()
-            self.chains[chain_name].publisher.heartbeat(30)
+    def process_IN_MOVED_TO(self, event):
+        """Process on closing after moving."""
+        self._fun(event.pathname)
 
-    def terminate(self):
-        """Terminate client chains."""
-        for chain in self.chains.values():
-            chain.stop()
-        LOGGER.info("Shutting down.")
-        print("Thank you for using pytroll/move_it_client."
-              " See you soon on pytroll.org!")
+
+class StatCollector(object):
+    """StatCollector class."""
+
+    def __init__(self, statfile):
+        """Initialize collector."""
+        self.statfile = statfile
+
+    def collect(self, msg, *args, **kwargs):
+        """Collect."""
+        with open(self.statfile, 'a') as fd:
+            fd.write(time.asctime() + " - " + str(msg) + "\n")
+
+
+def terminate(chains):
+    """Terminate client chains."""
+    for chain in six.itervalues(chains):
+        chain.stop()
+    LOGGER.info("Shutting down.")
+    print("Thank you for using pytroll/move_it_client."
+          " See you soon on pytroll.org!")
+    time.sleep(1)
+    sys.exit(0)
